#define GAIM_PLUGINS
#include "gaim.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <time.h>

#include <gtk/gtk.h>

#include "pixmaps/download.xpm"

#define SMARTEAR_XPM download_xpm
#define DEFAULT_NAME "(Default)"

struct sound_info {
    char type; // B,G
    char *name;
    char flags[5]; // AIMS
#define SOUNDFLAG_M	0
#define SOUNDFLAG_S	1
#define SOUNDFLAG_I	2
#define SOUNDFLAG_A	3
#define SOUNDFLAG_END	4
#define SOUNDFLAG_EMPTY "    "
    char *sound;
};
struct sound_arg
{
    struct gaim_connection *gc;
    struct message_info *msg;
    char *who;
    char event;
};

struct message_info {
    char *buddy;
    time_t last;
    time_t last_sent;
    struct sound_arg *timer_arg;
    gint timer;
};


void
on_clist_select_row                    (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data);

void
on_add_clicked                         (GtkButton       *button,
                                        gpointer         user_data);

void
on_change_clicked                      (GtkButton       *button,
                                        gpointer         user_data);

void
on_delete_clicked                      (GtkButton       *button,
                                        gpointer         user_data);

void
on_browse_clicked                      (GtkButton       *button,
                                        gpointer         user_data);

void
on_browse_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data);

gboolean
on_file_browse_destroy                 (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data);

void
on_test_clicked                        (GtkButton       *button,
                                        gpointer         user_data);

void
on_save_clicked                        (GtkButton       *button,
                                        gpointer         user_data);

void
on_apply_clicked                       (GtkButton       *button,
                                        gpointer         user_data);

gboolean
on_config_destroy                      (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data);

gint find_row_by_name(struct sound_info *entry);
struct sound_info *get_data_from_entries(void);
void set_entries_from_data(struct sound_info *data, gboolean use_name);
gint add_sound_entry(struct sound_info *entry);
gint clist_cmp(GtkCList *clist, const void *p1, const void *p2);
void init_config(void);
void clear_entries(void);
struct sound_info *sound_dup(struct sound_info *entry);
void sound_free(struct sound_info *entry);
gint sound_cmp(struct sound_info *entry1, struct sound_info *entry2);
void soundlist_free(void);
void smartear_save(void);
void smartear_load(void);
char *my_normalize(char* str);
void play_smart_sound(char *sound);

int message_delay = 60;
int focused_quiet = TRUE;
int smarterear_timers = TRUE;

gchar current_type = 0;
gchar prev_name[BUF_LEN] = "";

GSList *sounds_list = 0;
GSList *message_list = 0;

void *handle;

#include "smartear-interface.c"

void
on_clist_select_row                    (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    struct sound_info *entry;
    entry = (struct sound_info*)gtk_clist_get_row_data(clist, row);

    if (!entry) {
        g_warning("invalid row in clist_select (%d)", row);
        return;
    }

    set_entries_from_data(entry, FALSE);
}

void
on_add_clicked                         (GtkButton       *button,
                                        gpointer         user_data)
{
    struct sound_info *entry;
    char *p;

    gtk_clist_freeze(GTK_CLIST(clist));

    entry = get_data_from_entries();
    if (add_sound_entry(entry) < 0)
	sound_free(entry);
    else // added succesfully, empty the entries
	clear_entries();

    gtk_clist_thaw(GTK_CLIST(clist));
}

void
on_change_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    int row;
    struct sound_info *entry;

    gtk_clist_freeze(GTK_CLIST(clist));

    entry = get_data_from_entries();
    if ((row = find_row_by_name(entry)) >= 0)
	gtk_clist_remove(GTK_CLIST(clist), row);

    if (add_sound_entry(entry) < 0) {
	g_warning("couldn't change entry %s?", entry->name);
	sound_free(entry);
	gtk_clist_thaw(GTK_CLIST(clist));
	return;
    }

    if ((row=gtk_clist_find_row_from_data(GTK_CLIST(clist), entry)) >= 0)
	gtk_clist_select_row(GTK_CLIST(clist), row, 0);

    gtk_clist_thaw(GTK_CLIST(clist));
}

void
on_delete_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    int row;
    struct sound_info *entry;

    entry = get_data_from_entries();
    if (!(entry->type == 'B' && g_strcasecmp(entry->name, DEFAULT_NAME)==0)
	    && (row = find_row_by_name(entry)) >= 0) {
	gtk_clist_remove(GTK_CLIST(clist), row);
    }
    sound_free(entry);
}

void
on_browse_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    char *sound;
    if (file_browse)
	return;

    create_file_browse();
    sound = gtk_editable_get_chars(GTK_EDITABLE(entry_sound), 0, -1);
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(file_browse), sound);
    g_free(sound);
    gtk_widget_show(file_browse);
}

void
on_browse_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
    char *sound = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_browse));
    gtk_entry_set_text(GTK_ENTRY(entry_sound), sound);
}

gboolean
on_file_browse_destroy                 (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    file_browse = NULL;
    return FALSE;
}

void
on_test_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
    char *sound = gtk_editable_get_chars(GTK_EDITABLE(entry_sound), 0, -1);
    play_smart_sound(sound);
    g_free(sound);
}

void
on_save_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
    on_apply_clicked(button, user_data);
    smartear_save();
}


void
on_apply_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
    struct sound_info *entry;
    int row;

    soundlist_free();

    for (row = 0; ; row++) {
	if ((entry = (struct sound_info*)gtk_clist_get_row_data(GTK_CLIST(clist), row)) == 0)
	    break;
	sounds_list = g_slist_append(sounds_list, sound_dup(entry));
    }

    message_delay = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton_delay));
    focused_quiet = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_focusquiet));
    smarterear_timers = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_smarterear));
}

gboolean
on_config_destroy                      (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    config = NULL;
    if (file_browse)
	gtk_widget_destroy(file_browse);
	
    return FALSE;
}

void on_option_selected(GtkMenuShell *menu_shell, gpointer data)
{
    GtkWidget *menuitem;
    gchar newtype;
    struct sound_info tmp, *entry;

    menuitem = gtk_menu_get_active(GTK_MENU(menu_shell));
    newtype = g_list_index(menu_shell->children, menuitem)== 0 ? 'B' : 'G';

    entry = get_data_from_entries();
    strcpy(tmp.flags, entry->flags);
    tmp.sound = entry->sound;
    tmp.type = newtype;

    if (current_type == 'B' && newtype == 'G') {
	struct group *g;
	char *name = gtk_editable_get_chars(GTK_EDITABLE(entry_name), 0, -1);
	if ((g = find_group_by_buddy(NULL, name))) {
	    strncpy(prev_name, name, BUF_LEN);
	    tmp.name = my_normalize(g->name);
	    set_entries_from_data(&tmp, TRUE);
	}
	g_free(name);
    }
    else if (current_type == 'G' && newtype == 'B') {
	if (prev_name[0]) {
	    tmp.name = my_normalize(prev_name);
	    set_entries_from_data(&tmp, TRUE);
	}
	prev_name[0] = 0;
    }
    current_type = newtype;
}

gint find_row_by_name(struct sound_info *entry)
{
    struct sound_info *that;
    int row;

    for (row = 0; ; row++) {
	if ((that = (struct sound_info*)gtk_clist_get_row_data(GTK_CLIST(clist), row)) == 0)
	    break;
	if ((entry->type == 0 || entry->type == that->type)
		&& g_strcasecmp(entry->name, that->name)==0)
	    return row;
    }
    return -1;
}

struct sound_info *get_data_from_entries(void)
{
    struct sound_info *entry = g_new(struct sound_info, 1);
    char *tmp;

    tmp = gtk_editable_get_chars(GTK_EDITABLE(entry_name), 0, -1);
    entry->name = g_strdup(my_normalize(tmp));
    g_free(tmp);
    entry->sound = gtk_editable_get_chars(GTK_EDITABLE(entry_sound), 0, -1);
    entry->type = current_type;

    strcpy(entry->flags, SOUNDFLAG_EMPTY);
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_m)))
	entry->flags[SOUNDFLAG_M] = 'M';
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_s)))
	entry->flags[SOUNDFLAG_S] = 'S';
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_i)))
	entry->flags[SOUNDFLAG_I] = 'I';
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_a)))
	entry->flags[SOUNDFLAG_A] = 'A';

    return entry;
}

void set_entries_from_data(struct sound_info *entry, gboolean use_name)
{
    if (use_name) {
	GSList *lp;

	if ((lp = g_slist_find_custom(sounds_list, entry, (GCompareFunc)sound_cmp))) {
	    set_entries_from_data((struct sound_info*)lp->data, FALSE);
	    return;
	}
    }

    current_type = entry->type;

    gtk_option_menu_set_history(GTK_OPTION_MENU(optionmenu_type),
	 (entry->type=='B') ? 0 : 1);
    gtk_entry_set_text(GTK_ENTRY(entry_name), my_normalize(entry->name));
    gtk_entry_set_text(GTK_ENTRY(entry_sound), entry->sound);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_m),
	    (entry->flags[SOUNDFLAG_M] == 'M'));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_s),
	    (entry->flags[SOUNDFLAG_S] == 'S'));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_i),
	    (entry->flags[SOUNDFLAG_I] == 'I'));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_a),
	    (entry->flags[SOUNDFLAG_A] == 'A'));
}

gint add_sound_entry(struct sound_info *entry)
{
    char* data[4];
    int row;

    if (entry->name == 0 || *entry->name == 0)
	return -1;

    if (find_row_by_name(entry) >= 0)
	return -1;

    data[0] = entry->type == 'B' ? "Buddy" : "Group";
    data[1] = entry->name;
    data[2] = entry->flags;
    data[3] = entry->sound;

    row = gtk_clist_append(GTK_CLIST(clist), data);
    gtk_clist_set_row_data_full(GTK_CLIST(clist), row, entry,
	   (GtkDestroyNotify)sound_free);
    gtk_clist_sort(GTK_CLIST(clist));

    return 0;
}

gint clist_cmp(GtkCList *clist, const void *p1, const void *p2)
{
    GtkCListRow *row1 = (GtkCListRow*)p1;
    GtkCListRow *row2 = (GtkCListRow*)p2;
    struct sound_info *entry1 = (struct sound_info*)row1->data;
    struct sound_info *entry2 = (struct sound_info*)row2->data;

    if (!entry1)
	return -1;
    if (!entry2)
	return 1;

    if (entry1->type != entry2->type)
	return (entry1->type - entry2->type);
    return (g_strcasecmp(entry1->name, entry2->name));
}

void init_config(void)
{
    GSList *lp;

    if (config)
	return;

    current_type = 0;
    prev_name[0] = 0;

    config = create_config();

    gtk_signal_connect(
	GTK_OBJECT(gtk_option_menu_get_menu(GTK_OPTION_MENU(optionmenu_type))),
	"deactivate", GTK_SIGNAL_FUNC(on_option_selected), NULL);
    // set current_type
    on_option_selected(
	GTK_MENU_SHELL(gtk_option_menu_get_menu(GTK_OPTION_MENU(optionmenu_type))),
	NULL);

    gtk_clist_set_compare_func(GTK_CLIST(clist), (GtkCListCompareFunc)clist_cmp);
    gtk_clist_set_auto_sort(GTK_CLIST(clist), FALSE);

    gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton_delay),
	   (gfloat)message_delay);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_focusquiet),
	    focused_quiet);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_smarterear),
	    smarterear_timers);

    for (lp = sounds_list; lp; lp = g_slist_next(lp))
	add_sound_entry(sound_dup((struct sound_info*)lp->data));

    gtk_widget_show(config);
}

void clear_entries(void)
{
    struct sound_info tmp;

    tmp.type = 'B';
    tmp.name = "";
    strcpy(tmp.flags, SOUNDFLAG_EMPTY);
    tmp.sound = "";
    set_entries_from_data(&tmp, FALSE);

    gtk_clist_unselect_all(GTK_CLIST(clist));
}

struct sound_info *sound_dup(struct sound_info *entry)
{
    struct sound_info *edup = g_new(struct sound_info, 1);

    edup->type = entry->type;
    edup->name = g_strdup(my_normalize(entry->name));
    strcpy(edup->flags, entry->flags);
    edup->sound = g_strdup(entry->sound);

    return edup;
}

void sound_free(struct sound_info *entry)
{
    g_free(entry->name);
    g_free(entry->sound);
    g_free(entry);
}

gint sound_cmp(struct sound_info *entry1, struct sound_info *entry2)
{
    if (entry1->type != entry2->type
	    && entry1->type != 0 && entry2->type != 0)
	return entry1->type - entry2->type;
    return g_strcasecmp(entry1->name, entry2->name);
}

void soundlist_free(void)
{
    GSList *lp;
    for (lp = sounds_list; lp; lp = g_slist_next(lp))
	sound_free((struct sound_info*)lp->data);
    if (sounds_list) {
	g_slist_free(sounds_list);
	sounds_list = 0;
    }
}

char *my_normalize(char *input)
{
    static char buf[BUF_LEN];
    char *str, *beg;
    int i=0;

    g_return_val_if_fail((input != NULL), NULL);

    beg = str = g_strdup(input);

    while (*str && i <= BUF_LEN) {
	if (*str != ' ')
	    buf[i++] = *str;
	str++;
    }
    buf[i] = '\0';
    g_free(beg);

    return buf;

}

void smartear_save(void)
{
    char file[BUF_LEN] = "smartear.rc - you have no home dir";
    FILE *fp = NULL;
    GSList *lp;
    struct sound_info *entry;

    if (getenv("HOME")) {
	g_snprintf(file, sizeof(file), "%s/.gaim/smartear.rc", getenv("HOME"));
	fp = fopen(file, "w");
    }

    if (fp == NULL) {
	g_warning("couldn't open %s\n", file);
	return;
    }

    fprintf(fp, "smarterear_timers %d\n", smarterear_timers);
    fprintf(fp, "delay %d\n", message_delay);
    fprintf(fp, "focused_quiet %d\n", focused_quiet);

    for (lp = sounds_list; lp; lp = g_slist_next(lp)) {
	entry = (struct sound_info*)lp->data;
	fprintf(fp, "%c {%s} {%s} {%s}\n",
	    entry->type, entry->name, entry->flags, entry->sound);
    }

    fclose(fp);
}

void smartear_load(void)
{
    char buf[BUF_LEN];
    char name[BUF_LEN];
    char flags[BUF_LEN];
    char file[BUF_LEN] = "smartear.rc - you have no home dir";
    FILE *fp = NULL;
    struct sound_info *entry;
    char type;
    gboolean has_default = FALSE;

    if (getenv("HOME")) {
	g_snprintf(file, sizeof(file), "%s/.gaim/smartear.rc", getenv("HOME"));
	fp = fopen(file, "r");
    }

    if (fp == NULL) {
	g_warning("couldn't open %s\n", file);
	entry = g_new(struct sound_info, 1);
	entry->type = 'B';
	entry->name = g_strdup(DEFAULT_NAME);
	strcpy(entry->flags, SOUNDFLAG_EMPTY);
	entry->flags[SOUNDFLAG_M] = 'M';
	entry->sound = g_strdup("");

	sounds_list = g_slist_append(sounds_list, entry);
	return;
    }

    // FIXME: This is a hack.. do this right.
    fgets(buf, sizeof(buf), fp);
    if(strncmp(buf, "smarterear_timers", strlen("smarterear_timers")) == 0) {
        if (sscanf(buf, "smarterear_timers %d", &smarterear_timers) != 1)
            g_warning("smartear: syntax error: %s", buf);
        fgets(buf, sizeof(buf), fp);
    }

    if (sscanf(buf, "delay %d", &message_delay) != 1)
	g_warning("smartear: syntax error: %s", buf);

    fgets(buf, sizeof(buf), fp);

    if (sscanf(buf, "focused_quiet %d", &focused_quiet) != 1)
	g_warning("smartear: syntax error: %s", buf);
    

    while ((fgets(buf, sizeof(buf), fp))) {
	if (sscanf(buf, "%c {%[^}]} {%[^}]} {%[^}]}", &type, name, flags, file) != 4) {
	    g_warning("smartear: syntax error: %s", buf);
	    continue;
	}

	if (type != 'G' && type != 'B') {
	    g_warning("smartear: no such type: %c", type);
	    continue;
	}

	if (!strchr("M ", flags[SOUNDFLAG_M]) ||
	    !strchr("S ", flags[SOUNDFLAG_S]) ||
	    !strchr("I ", flags[SOUNDFLAG_I]) ||
	    !strchr("A ", flags[SOUNDFLAG_A])) {
	    g_warning("smartear: invalid flags: %s", flags);
	    strcpy(flags, SOUNDFLAG_EMPTY);
	}

	entry = g_new(struct sound_info, 1);
	entry->type = type;
	entry->name = g_strdup(my_normalize(name));
	strncpy(entry->flags, flags, SOUNDFLAG_END);
	entry->flags[SOUNDFLAG_END] = 0;
	entry->sound = g_strdup(file);

	sounds_list = g_slist_append(sounds_list, entry);

	if (g_strcasecmp(entry->name, DEFAULT_NAME) == 0)
	    has_default = TRUE;
    }
    if (!has_default) {
	entry = g_new(struct sound_info, 1);
	entry->type = 'B';
	entry->name = g_strdup(DEFAULT_NAME);
	strcpy(entry->flags, SOUNDFLAG_EMPTY);
	entry->flags[SOUNDFLAG_M] = 'M';
	entry->sound = g_strdup("");

	sounds_list = g_slist_append(sounds_list, entry);
    }

    fclose(fp);
}

void on_smarticon_clicked(GtkWidget *item, struct conversation *conv)
{
    struct sound_info tmp, *entry;
    int row;

    if (!config)
	init_config();

    gtk_clist_unselect_all(GTK_CLIST(clist));

    tmp.type = 'B';
    tmp.name = my_normalize(conv->name);
    strcpy(tmp.flags, SOUNDFLAG_EMPTY);
    tmp.flags[SOUNDFLAG_M] = 'M';
    tmp.sound = "(default)";

    set_entries_from_data(&tmp, TRUE);
}

void play_smart_sound(char *sound)
{
    struct sound_info tmp, *entry;
    GSList *lp;

    if (!sound || !*sound)
	return;

    // sound aliases: mostly so you can put (Default) for sound, and it'll
    // play that one
    tmp.type = 0;
    tmp.name = sound;
    if ((lp = g_slist_find_custom(sounds_list, &tmp, (GCompareFunc)sound_cmp))) {
	entry = (struct sound_info*)lp->data;
	play_smart_sound(entry->sound);
    }
    else
	play_file(sound);
}

void play_matching_sound(struct gaim_connection *gc, char *who, char event)
{
    GSList *lp;
    struct sound_info *sinfo;
    struct group *g;
    char *sound = NULL;
    char *name = g_strdup(my_normalize(who));

    for (lp = sounds_list; lp; lp = g_slist_next(lp)) {
	sinfo = (struct sound_info*)lp->data;
	if (!strchr(sinfo->flags, event))
	    continue;

	if (sinfo->type == 'B' && g_strcasecmp(name, sinfo->name) == 0) {
	    sound = sinfo->sound;
	    break; // found a buddy match.. this takes precedence, so stop
	}
	if (sinfo->type == 'G' && (g = find_group_by_buddy(gc, name)) &&
		g_strcasecmp(my_normalize(g->name), sinfo->name) == 0) {
	    sound = sinfo->sound; // keep going... buddy overrides group
	}
	if (sinfo->type == 'B' && g_strcasecmp(sinfo->name, DEFAULT_NAME) == 0) {
	    if (!sound)
		sound = sinfo->sound; // keep going.. other 2 override default
	}
    }
    play_smart_sound(sound);

    g_free(name);
}


gint play_sound_timer_hook(gpointer data)
{
    struct message_info *msg = data;

    if(!msg->timer_arg) {
        g_print("smartear: timer called without being active!\n");
        return 0;
    }
    
    play_matching_sound(msg->timer_arg->gc, msg->buddy, msg->timer_arg->event);
    
    g_free(msg->timer_arg);
    msg->timer_arg = NULL;

    return 0;
}

void on_im_recv(struct gaim_connection *gc, char **who, char **what, gint32 *flags, void *junk)
{
    struct message_info *msg;
    GSList *lp;
    GList *dlp;
    time_t now = time(0);
    char *name = g_strdup(my_normalize(*who));
    struct conversation *conv = find_conversation(*who);

    // FIXME: message list is never trimmed. However, if we DO trim it,
    // we need to consider race conditions on the msg field of the timer_hook
    for (lp = message_list; lp; lp = g_slist_next(lp)) {
	msg = (struct message_info*)lp->data;
	if (g_strcasecmp(name, msg->buddy) == 0) {
      
            // only install a timer if there's no active timers and we've just 
            // sent a message. If we haven't just sent a message, then 
            // a beep will go off anyways when the IM arrives.
            if(smarterear_timers && !msg->timer_arg
                    && msg->last_sent + message_delay > now) {
                struct sound_arg *arg = g_new0(struct sound_arg, 1);

                arg->gc = gc;
                arg->event = 'M';
                msg->timer_arg = arg;
                
                gtk_timeout_add(message_delay*1000, play_sound_timer_hook, msg);
            }

            // If this is the guy.. If it hasn't been the delay seconds, 
            // then don't install a new timer, otherwise add it..
            if (msg->last + message_delay > now && conv) {
                g_free(name);
                return;
            }
	    msg->last = now;
	    break; // else, we're done
	}
    }

    if (!lp) { // not in list
	msg = g_new0(struct message_info, 1);
	msg->buddy = g_strdup(name);
	msg->last = now;
        msg->timer_arg = NULL;

	message_list = g_slist_append(message_list, msg);
    }

    if (conv && GTK_WIDGET_HAS_FOCUS(conv->entry) && focused_quiet) {
	g_free(name);
	return; // don't play sounds for the focused convo
    }

    if (!conv || conv->makesound)
	play_matching_sound(gc, name, 'M');

    g_free(name);
}

void on_im_send(struct gaim_connection *gc, char *who, char **what, void *junk)
{
    struct message_info *msg;
    GSList *lp;
    time_t now = time(0);
    char *name = g_strdup(my_normalize(who));

    struct conversation *conv = find_conversation(who);

    for (lp = message_list; lp; lp = g_slist_next(lp)) {
        msg = (struct message_info*)lp->data;
	if (g_strcasecmp(name, msg->buddy) == 0) {
	    msg->last = now;
	    msg->last_sent = now;
            if(msg->timer_arg) {
                gtk_timeout_remove(msg->timer);
                g_free(msg->timer_arg);
                msg->timer_arg = NULL;
            }
	    break; // else, we're done
	}
    }
    if (!lp) { // not in list
	msg = g_new0(struct message_info, 1);
	msg->buddy = g_strdup(name);
	msg->last = now;
        msg->last_sent = now;
        msg->timer_arg = NULL;

	message_list = g_slist_append(message_list, msg);
    }

    g_free(name);
}

void on_buddy_signon(struct gaim_connection *gc, char *who, void *junk)
{
    play_matching_sound(gc, who, 'S');
}

void on_buddy_back(struct gaim_connection *gc, char *who, void *junk)
{
    play_matching_sound(gc, who, 'A');
}

void on_buddy_unidle(struct gaim_connection *gc, char *who, void *junk)
{
    play_matching_sound(gc, who, 'I');
}

static void clear_timers(struct gaim_connection *gc, void *m) {
    struct message_info *msg;
    GSList *lp;

    for (lp = message_list; lp; lp = g_slist_next(lp)) {
         msg = (struct message_info*)lp->data;

         if(msg->timer_arg) {
             gtk_timeout_remove(msg->timer);
             g_free(msg->timer_arg);
             msg->timer_arg = NULL;
         }
    }
}

void on_new_conv(char *who, struct conversation *junk)
{
    GdkPixmap *icon_i;
    GdkBitmap *mask;
    GtkWidget *win, *icon_p, *icon;
    struct conversation *conv;

    conv = junk;

    if (!conv)
	conv = find_conversation(who);
    if (!conv)
	return;

    win = conv->window;
    icon_i = gdk_pixmap_create_from_xpm_d(
	   win->window, &mask, &win->style->white, SMARTEAR_XPM);
    icon_p = gtk_pixmap_new(icon_i, mask);
    gtk_widget_show(icon_p);
    gdk_bitmap_unref(mask);

    icon = gtk_toolbar_append_item(GTK_TOOLBAR(conv->toolbar),
	    NULL, _("Add SmartEar Buddy"), _("SmartEar"),
	    icon_p, GTK_SIGNAL_FUNC(on_smarticon_clicked), conv);

    gtk_object_set_data(GTK_OBJECT(conv->toolbar), "smarticon", icon);

    if (misc_options & OPT_MISC_COOL_LOOK)
	gtk_button_set_relief(GTK_BUTTON(icon), GTK_RELIEF_NONE);

    gdk_pixmap_unref(icon_i);
}

char *gaim_plugin_init(GModule *h)
{
    GList *lp;

    handle = h;

    smartear_load();

    gaim_signal_connect(handle, event_im_recv, on_im_recv, NULL);
    gaim_signal_connect(handle, event_im_send, on_im_send, NULL);
    gaim_signal_connect(handle, event_buddy_signon, on_buddy_signon, NULL);
    gaim_signal_connect(handle, event_buddy_back, on_buddy_back, NULL);
    gaim_signal_connect(handle, event_buddy_unidle, on_buddy_unidle, NULL);
    gaim_signal_connect(handle, event_new_conversation, on_new_conv, NULL);
    gaim_signal_connect(handle, event_signoff, clear_timers, NULL);

    for (lp = conversations; lp; lp = g_list_next(lp)) {
	struct conversation *conv = (struct conversation*)lp->data;
	on_new_conv(conv->name, conv);
    }

    return NULL;
}

char *gaim_plugin_config()
{
    init_config();

    return NULL;
}   

void gaim_plugin_remove(GModule *h)
{
    GSList *lp;
    GList *dlp;
    struct message_info *r;

    gtk_widget_destroy(config);
    config = NULL;

    soundlist_free();

    for (lp = message_list; lp; lp = g_slist_next(lp)) {
	r = (struct message_info*)lp->data;
	g_free(r->buddy);
	g_free(r);
    }
    g_slist_free(message_list);

    for (dlp = conversations; dlp; dlp = g_list_next(dlp)) {
	struct conversation *conv = (struct conversation*)dlp->data;
	GtkWidget *smarticon = 
	    (GtkWidget*)gtk_object_get_data(GTK_OBJECT(conv->toolbar), "smarticon");
	if (smarticon) {
	    gtk_object_remove_data(GTK_OBJECT(conv->toolbar), "smarticon");
	    gtk_container_remove(GTK_CONTAINER(conv->toolbar), smarticon);
	    gtk_widget_destroy(smarticon);
	}
    }
}

char *name()
{
    return "SmartEar sound plugin v0.0.1";
}

char *description()
{
    return
"Tired of rushing to the computer when you hear an IM noise, "
"expecting an urgent message, only to find out it's your jerk "
"cousin with news about his driveway basketball game victory?  "
"Sick of being interrupted by people you don't want to "
"talk to?  You need this plugin.  With it, you can assign different "
"sounds to play for different buddies or whole groups of buddies.  "
"SmartEar allows you to opt to play sounds when a buddy sends you an "
"IM, signs on, returns from away or idle, or any combination of these, so "
"you'll know by the sound what the important people are doing.  "
"You'll never have to talk to your relatives-in-law again.";
}
